module ListOps (unique, uniquely, (~+~), (~=~), (~-~), (~<~)) where

-- (~+~) adds two lists. 
-- Elements from the right list will only be added if they're not already
-- contained in the left list. The function is defined recursively on the right list.
-- (calling the function returns a new list and leaves the original lists unchanged)
(~+~) :: (Eq a) => [a] -> [a] -> [a]
(~+~) xs [] = xs
(~+~) xs (y:ys)
	| y `elem` xs = xs ~+~ ys
	| otherwise   = (y:xs) ~+~ ys

-- (~-~) subtracts two lists.
-- Elements will be removed from the left list if they're in the right list.
(~-~) :: (Eq a) => [a] -> [a] -> [a]
(~-~) [] ys = []
(~-~) xs ys = [x | x <- xs, not (x `elem` ys)]

-- (~=~) tests whether lists are equal or equivalent.
-- Under this function, lists are the same if they each contain the same elements.
-- Note that a list with duplicates will be equivalent to a list without.
(~=~) :: (Eq a) => [a] -> [a] -> Bool
(~=~) xs ys = all (flip elem ys) xs && all (flip elem xs) ys

-- (~<~) tests whether a list is a subset of a list
(~<~) :: (Eq a) => [a] -> [a] -> Bool
(~<~) [] ys = True
(~<~) (x:xs) ys
	| x `elem` ys = xs ~<~ ys
	| otherwise   = False

-- Calls a helper function to build a list of unique elements
unique :: (Eq a) => [a] -> [a]
unique xs = uniq xs []

-- Recursively builds a list without duplicates.
-- The argument on the right is used to build the new list
uniq :: (Eq a) => [a] -> [a] -> [a]
uniq [] ys = ys
uniq (x:xs) ys
	| any (x ==) ys = uniq xs ys
	| otherwise     = uniq xs (x:ys) 

-- A version of the above function but for lists of lists
-- where list equivalence is used instead of equality
uniqL :: (Eq a) => [[a]] -> [[a]] -> [[a]]
uniqL [] ys = ys
uniqL (x:xs) ys
	| any (x ~=~) ys = uniqL xs ys
	| otherwise		 = uniqL xs (x:ys)

-- Calls the helper function for the list of lists
uniquely :: (Eq a) => [[a]] -> [[a]]
uniquely xs = uniqL xs []
