module Thickness (execThickness, Flag (Short,Long,Extra)) where

import Graph
import Joins
import ListOps
import Control.Monad.State
import Control.Monad.Writer
import Control.Monad.Identity
import Control.Monad.Reader
import Control.Parallel.Strategies
import Control.Parallel (par,pseq)
import Control.DeepSeq

type Results a = WriterT String (ReaderT (Graph,Flag) (StateT ([Graph],Maybe Bool) Identity)) a

runResults True flag n graph = runIdentity (runStateT (runReaderT (runWriterT $ thickness algorithm' n) (graph,flag)) ([],Nothing))
runResults False flag n graph = runIdentity (runStateT (runReaderT (runWriterT $ thickness algorithm n) (graph,flag)) ([],Nothing))

execThickness :: Bool -> Flag -> Int -> Graph -> String
execThickness b f n g = if log == "" then maybe "n/a" show result ++ "\n" else maybe "n/a" show result ++ ": " ++ log
	where (result,log) = fst $ runResults b f n g

data Flag = Short | Long | Extra

instance Eq Flag where
	Short == Short = True
	Long == Long = True
	Extra == Extra = True
	_ == _ = False

startM :: Results ([Graph],JoinState)
startM = do
	(graph,flag) <- ask
	if flag == Extra then tell "Generating first batch of blocks...\n" else return ()
	return . runIdentity $ runStateT (runReaderT (start) graph) ([],[])

nextM :: JoinState -> Results ([Graph],JoinState)
nextM gen = do
	(graph,flag) <- ask
	if flag == Extra then tell "Generating next batch of blocks...\n" else return ()
	return . runIdentity $ runStateT (runReaderT (next) graph) gen

orderZero :: ([Graph],JoinState) -> ([Graph],JoinState)
orderZero (blocks,j) = (withStrategy (parList rpar) . filter ((2<=) . length . filter notComplete . joinComponents) $ blocks,j)

(>~>) :: (JoinState -> Results ([Graph],JoinState)) -> (JoinState -> Results ([Graph],JoinState)) -> JoinState -> Results ([Graph],JoinState)
a >~> b = \j -> do 
	(blocks2,j2) <- a j
	(blocks3,j3) <- b j2
	return (blocks2 ++ blocks3 `using` (parList rpar),j3)

thickness :: (Int -> Results (Maybe Int)) -> Int -> Results (Maybe Int)
thickness function n = do
	(graph,flag) <- ask
	if runReader (isConnected) graph == False
	then do
		if flag == Extra then tell "Graph is disconnected.\n" else return ()
		return Nothing
	else do
		let m = max (n-1) 0
		if m == 0 
		then startM >>= return . orderZero >>= refine function
		else return ([],[]) >>= foldl (>~>) (\_ -> startM) (replicate m nextM) 
			>>= return . orderZero >>= refine function

refine :: (Int -> Results (Maybe Int)) -> ([Graph],JoinState) -> Results (Maybe Int)
refine function (blocks,gen) = do
	(_,flag) <- ask
	put (blocks,Nothing)
	case flag of
		Short -> return ()
		Long -> do
			if length blocks == 1
			then tell $ "At thickness 0, there is 1 block.\n"
			else tell $ "At thickness 0, there are " ++ (show $ length blocks) ++ " blocks.\n"
		Extra -> do
			if length blocks == 1
			then tell $ "At thickness 0, there is 1 block: " ++ show blocks ++ ".\n"
			else tell $ "At thickness 0, there are " ++ (show $ length blocks) ++ " blocks: "
				++ show blocks ++ ".\n"
	test <- function 0
	if test /= Nothing
	then do
		if flag == Extra then tell "Success!\n" else return ()
		return test
	else do
		if flag == Extra then tell "Failure!\n" else return ()
		(nextblocks,nextgen) <- return . orderZero =<< nextM gen
		if nextblocks == []
		then do
			case flag of
				Short -> return ()
				Long -> tell "Failure!\n"
				Extra -> tell "No new blocks to try.\n"
			return Nothing
		else do 
			case flag of
				Short -> return ()
				Long -> tell "Failed; trying again...\n"
				Extra -> tell "Trying again...\n"
			refine function $ orderZero (blocks ~+~ nextblocks,nextgen)

algorithm :: Int -> Results (Maybe Int)
algorithm n = do
	(blocks,last) <- get
	if blocks == [] then return Nothing else do
		(graph,_) <- ask
		if graph `elem` blocks then return $ Just n else do
			case last of
				Just True -> do
					m <- doCones n
					if n == m 
					then do
						o <- doUnions n
						if o == n then return Nothing else algorithm o
					else algorithm m
				_ -> do
					m <- doUnions n
					if n == m
					then do
						o <- doCones n
						if o == n then return Nothing else algorithm o
					else algorithm m

algorithm' :: Int -> Results (Maybe Int)
algorithm' n = do
	(blocks,_) <- get
	if blocks == [] then return Nothing else do
		(graph,_) <- ask
		if graph `elem` blocks then return $ Just n else do
			m <- doUnions n
			if m == n
			then do
				o <- doCones n
				if n == o then return Nothing else algorithm' o
			else algorithm m

doCones :: Int -> Results Int
doCones n = do
	(graph,flag) <- ask
	(blocks,_) <- get
	let newBlocks = parMap (rparWith rdeepseq) (\g -> runReader (cone g) graph) blocks
	if newBlocks ~=~ blocks
	then return n
	else do
		case flag of
			Short -> return ()
			Long -> do
				if length newBlocks == 1
				then tell $ "At thickness " ++ show (n + 1) ++ ", after coning, there is 1 block.\n"
				else tell $ "At thickness " ++ show (n + 1) ++ ", after coning, there are "
					++ show (length newBlocks) ++ " blocks.\n"
			Extra -> do
				if length newBlocks == 1
				then tell $ "At thickness " ++ show (n + 1) ++ ", after coning, there is 1 block: "
					++ show newBlocks ++ ".\n"
				else tell $ "At thickness " ++ show (n + 1) ++ ", after coning, there are "
					++ show (length newBlocks) ++ " blocks: " ++ show newBlocks ++ ".\n"
		put (newBlocks,Just False)
		return $ n+1

doUnions :: Int -> Results Int
doUnions n = do
	(graph,flag) <- ask
	(blocks,_) <- get
	let newBlocks = runReader (unionhelper blocks) graph
	if newBlocks ~=~ blocks
	then return n
	else do
		case flag of
			Short -> return ()
			Long -> do
				if length newBlocks == 1
				then tell $ "At thickness " ++ show (n + 1) ++ ", after taking unions, there is 1 block.\n"
				else tell $ "At thickness " ++ show (n + 1) ++ ", after taking unions, there are "
					++ show (length newBlocks) ++ " blocks.\n"
			Extra -> do
				if length newBlocks == 1
				then tell $ "At thickness " ++ show (n + 1) ++ ", after taking unions, there is 1 block: "
					++ show newBlocks ++ ".\n"
				else tell $ "At thickness " ++ show (n + 1) ++ ", after taking unions, there are "
					++ show (length newBlocks) ++ " blocks: " ++ show newBlocks ++ ".\n"
		put (newBlocks,Just True)
		return $ n+1

notComplete :: Graph -> Bool
notComplete graph
	| (length . vertices) graph <= 1 = False
	| otherwise = any (not . (`elem` edges graph) . edge)
		[(v,w) | v <- vertices graph, w <- vertices graph, v < w]

overlap :: Graph -> Graph -> WithGraph Bool
overlap j k = do
	graph <- ask
	return . notComplete . (graph ~:) $ [v | v <- vertices j, v `elem` (vertices k)]

n :: Graph -> Vertex -> [Vertex]
n graph vertex = [v | v <- vertices graph, edge(v,vertex) `elem` edges graph]

intersects :: (Eq a) => [a] -> [a] -> Bool
x `intersects` y = any (`elem` x) y

cone :: Graph -> WithGraph Graph
cone block = do	
	graph <- ask
	let cs = runReader (cliques) . (graph ~:) . filter
			(notComplete . (graph ~:) . (filter (`elem` vertices block)) . (n graph))
			$ (vertices graph) ~-~ (vertices block)
	if cs == []
	then return block
	else do
		let m = maximum $ map length cs
		toGraph . (vertices block ~+~) . head . filter ((m==) . length) $ cs

unionhelper :: [Graph] -> WithGraph [Graph]
unionhelper bs = do
	graph <- ask
	return . withStrategy (parList rpar)
		=<< mapM (toGraph . foldl (~+~) [] . map vertices) =<< foldM (helper) [] bs

helper :: [[Graph]] -> Graph -> WithGraph [[Graph]]
helper blocks block = do
	graph <- ask
	if any ((~=~ vertices graph) . foldl (~+~) [] . map vertices) blocks 
	then return blocks
	else do
		head <- return . withStrategy (parList $ rparWith rdeepseq) . foldl (~+~) [block] .
			filter (any (\g -> runIdentity $ runReaderT (overlap block g) graph)) $ blocks
		tail <- return . withStrategy (parList $ rparWith rdeepseq) =<< 
			filterM (return . not . any (\g -> runIdentity $ runReaderT (overlap block g) graph)) blocks
		return $ (force head `par` force tail) `pseq` (head : tail)
